/* $Id: nfs.C,v 1.17 2001/10/12 21:27:28 dm Exp $ */

/*
 *
 * Copyright (C) 2001 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "nfs.h"

const u_int maxfhiv = 512;
AUTH *cryptfs::ao = authopaque_create ();
tailq<cryptfs::fhiv, &cryptfs::fhiv::lrulink> cryptfs::ivlruq;
u_int cryptfs::fhiv::nfhiv;

inline void
xorblock (char *dest, const char *src)
{
  for (u_int i = 0; i < blocksize; i++)
    dest[i] ^= src[i];
}

u_int64_t
cryptfs::fhiv::attr2iv (const fattr3exp *f)
{
  /* Note:  We make the iv a deterministic function of the attributes
   * in case two clients simultaneously perform UNCHECKED creates of
   * the same file.  We could alternatively hash the file handle, but
   * in the case of SFS, different clients will have different file
   * handles (because sfsrwcd encrypts file handles to make it
   * slightly resistant to forged "loopback" traffic). */
  u_char buf[sha1::hashsize];
  puthyper (buf, f->fileid);
  puthyper (buf + 8, f->fsid);
  sha1_hash (buf, buf, 16);
  return gethyper (buf);
}

void
cryptfs::fhiv::init ()
{
  cryptfs::ivlruq.insert_tail (this);
  cfs->ivtab.insert (this);
  nfhiv++;
  while (nfhiv > maxfhiv && cfs->ivlruq.first->stable)
    delete cfs->ivlruq.first;
}

void
cryptfs::fhiv::fetchcb (ref<cryptfs> hold, ref<read3res> resp, clnt_stat stat)
{
  if (stat || resp->status || resp->resok->data.size () < blocksize)
    err = true;
  else {
    char *buf = resp->resok->data.base ();
    cfs->decrypt (-blocksize, 0, buf, buf);
    if (gethyper (buf))
      err = true;
    else
      iv = gethyper (buf + 8);
  }
  wake ();
}

void
cryptfs::fhiv::createcb (ref<cryptfs> hold, ref<write3res> resp,
			 clnt_stat stat)
{
  if (stat || resp->status)
    err = true;
  wake ();
}

void
cryptfs::fhiv::wake ()
{
  stable = valid = true;
  touch ();

  typedef vec<cbv> ivw_t;
  if (ivw_t *p = cfs->ivwait[fh]) {
    ivw_t v;
    p->swap (v);
    cfs->ivwait.remove (fh);
    while (!v.empty ())
      (*v.pop_front ()) ();
  }
}

cryptfs::fhiv::fhiv (cryptfs *fs, const nfs_fh3 &f, bool isdir)
  : cfs (fs), fh (f), iv (0), valid (false), stable (false), err (false)
{
  assert (!isdir);
  init ();

  read3args arg;
  arg.file = fh;
  arg.offset = 0;
  arg.count = blocksize;
  ref<read3res> resp = New refcounted<read3res>;
  cfs->call (NFSPROC3_READ, &arg, resp,
	     wrap (this, &cryptfs::fhiv::fetchcb, mkref (cfs), resp),
	     cfs->ao);
}

cryptfs::fhiv::fhiv (cryptfs *fs, const nfs_fh3 &f, const fattr3exp *fa)
  : cfs (fs), fh (f), iv (attr2iv (fa)),
    valid (true), stable (false), err (false)
{
  init ();
  write3args arg;
  arg.file = fh;
  arg.offset = 0;
  arg.count = blocksize;
  arg.stable = DATA_SYNC;
  arg.data.setsize (blocksize);
  puthyper (arg.data.base (), 0);
  puthyper (arg.data.base () + 8, iv);
  cfs->encrypt (-blocksize, 0, arg.data.base (), arg.data.base ());
  ref<write3res> resp = New refcounted<write3res>;
  cfs->call (NFSPROC3_WRITE, &arg, resp,
	     wrap (this, &cryptfs::fhiv::createcb, mkref (cfs), resp),
	     cfs->ao);
}

cryptfs::fhiv::~fhiv ()
{
  cryptfs::ivlruq.remove (this);
  cfs->ivtab.remove (this);
  nfhiv--;
}

void
cryptfs::fhiv::touch ()
{
  cryptfs::ivlruq.remove (this);
  cryptfs::ivlruq.insert_tail (this);
}

void
cryptfs::encrypt (u_int64_t offset, u_int64_t iv, char *out, const char *in)
{
  assert (!(offset & 15));
  char buf[16];
  puthyper (buf, iv);
  puthyper (buf + 8, offset);
  fskey.encipher_bytes (buf, buf);
  xorblock (buf, in);
  fskey.encipher_bytes (out, buf);
}

void
cryptfs::decrypt (u_int64_t offset, u_int64_t iv, char *out, const char *in)
{
  assert (!(offset & 15));
  char buf[16];
  puthyper (buf, iv);
  puthyper (buf + 8, offset);
  fskey.encipher_bytes (buf, buf);
  fskey.decipher_bytes (out, in);
  xorblock (out, buf);
}

str
cryptfs::encodename (u_int64_t diriv, str name)
{
  static str zeros ("\0", 2);
  assert (name.len ());
  if (name == "." || name == "..")
    return name;
  name = name << zeros;
  mstr b (name.len () + blockmask & ~blockmask);
  memcpy (b, name, name.len ());
  bzero (b + name.len (), b.len () - name.len ());

  fskey.encipher_bytes (b);
  for (char *p = b.cstr () + blocksize, *e = b.cstr () + b.len ();
       p < e; p += blocksize) {
    xorblock (p, p - blocksize);
    fskey.encipher_bytes (p);
  }

  /* Now encrypt backwards in CBC mode, so all output bits depend on
   * all input bits. */
  for (char *p = b.cstr () + b.len () - 2*blocksize; p >= b; p -= blocksize) {
    xorblock (p, p + blocksize);
    fskey.encipher_bytes (p);
  }

  return armor64A (b);
}

str
cryptfs::decodename (u_int64_t diriv, str name)
{
  if (name == "." || name == "..")
    return name;
  name = dearmor64A (name);
  if (name.len () & blockmask)
    return NULL;
  mstr b (name.len ());
  memcpy (b, name, name.len ());

  for (char *p = b, *e = b.cstr () + b.len () - blocksize;
       p < e; p += blocksize) {
    fskey.decipher_bytes (p);
    xorblock (p, p + blocksize);
  }

  for (char *p = b.cstr () + b.len () - blocksize; p > b; p -= blocksize) {
    fskey.decipher_bytes (p);
    xorblock (p, p - blocksize);
  }
  fskey.decipher_bytes (b);
  *(b.cstr () + b.len ()) = '\0';
  u_int blen = strlen (b.cstr ());
  if (b.len () - blen < 2)
    return NULL;
  for (u_int i = blen; i < b.len (); i++)
    if (b[i])
      return NULL;
  b.setlen (blen);
  return b;
}

int
cryptfs::getiv (nfs_fh3 &fh, bool stable, const authunix_parms *aup,
		cbv::ptr cb, u_int64_t *ivp)
{
  fhiv *fp = ivtab[fh];
  if (fp && fp->valid && (fp->stable || !stable)) {
    if (ivp)
      *ivp = fp->iv;
    fp->touch ();
    return fp->err ? -1 : 1;
  }

  if (cb) {
    vec<cbv> *p = ivwait[fh];
    if (!p) {
      ivwait.insert (fh);
      p = ivwait[fh];
    }
    p->push_back (cb);
  }
  
  if (!fp) {
    authopaque_set (ao, aup);
    vNew fhiv (this, fh, false);
  }
  return 0;
}

void
cryptfs::fillpoa (nfscall *nc, const nfs_fh3 *fhp,
		  post_op_attr *poa, aclnt_cb cb)
{
  ref<getattr3res> ares = New refcounted<getattr3res>;
  call (NFSPROC3_GETATTR, fhp, ares,
	wrap (mkref (this), &cryptfs::fillpoa_1, nc, poa, ares, cb), ao);
}
void
cryptfs::fillpoa_1 (nfscall *nc, post_op_attr *poa, ref<getattr3res> ares,
		    aclnt_cb cb, clnt_stat stat)
{
  if (stat)
    (*cb) (stat);
  else if (ares->status)
    nc->error (ares->status);
  else {
    poa->set_present (true);
    *poa->attributes = *ares->attributes;
    (*cb) (RPC_SUCCESS);
  }
}

void
cryptfs::reply (nfscall *nc, clnt_stat stat)
{
  if (stat) {
    warn << "NFS server: " << stat << "\n";
    nc->reject (SYSTEM_ERR);
    return;
  }
  else if (nc->proc () == NFSPROC3_NULL) {
    nc->reply (NULL);
    return;
  }

  nfs3_exp_enable (nc->proc (), nc->getvoidres ());

  xattrvec xv;
  nfs3_getxattr (&xv, nc->proc (), nc->getvoidarg (), nc->getvoidres ());
  for (xattr *x = xv.base (); x < xv.lim (); x++)
    if (x->fattr) {
      x->fattr->size = fixsize (x->fattr->type, x->fattr->size);
      if (x->wattr)
	x->wattr->size = fixsize (x->fattr->type, x->wattr->size);
    }
    else if (x->wattr)
      // without ftype, don't know how to adjust size, so nuke it
      x->wdata->before.set_present (false);

  nfs3_exp_disable (nc->proc (), nc->getvoidres ());

  if (nc->proc () != NFSPROC3_NULL && !*nc->template getres<nfsstat3> ())
    switch (nc->proc ()) {
    case NFSPROC3_WRITE:
      {
	write3args *arg = nc->template getarg<write3args> ();
	write3res *res = nc->template getres<write3res> ();
	res->resok->count -= arg->offset & blockmask;
	if (res->resok->count > arg->count)
	  res->resok->count = arg->count;
	break;
      }
    case NFSPROC3_READLINK:
      {
	readlink3res *res = nc->template getres<readlink3res> ();
	if (str link = decodename (0, res->resok->data))
	  res->resok->data = substr (link, 8);
	else
	  res->resok->data = "";
	break;
      }
    case NFSPROC3_READDIR:
      if (!fixreaddir<NFSPROC3_READDIR> (nc))
	return;
      break;
    case NFSPROC3_READDIRPLUS:
      if (!fixreaddir<NFSPROC3_READDIRPLUS> (nc))
	return;
      break;
    case NFSPROC3_ACCESS:
      if (aid != sfsaid_nobody && aid != aup2aid (nc->getaup ())
	  && (!nc->getaup () || (nc->getaup ()->aup_uid)))
	nc->template getres<access3res> ()->resok->access = 0;
      break;
    }

  nc->reply (nc->getvoidres ());
}

void
cryptfs::reply_create (nfscall *nc, ptr<lookup3res> lres,
		       ptr<getattr3res> ares, clnt_stat stat)
{
  if (stat) {
    reply (nc, stat);
    return;
  }

  create3args *argp = nc->template getarg<create3args> ();
  diropres3 *resp = nc->template getres<diropres3> ();
  authopaque_set (ao, nc->getaup ());

  if (!resp->resok->obj.present) {
    if (!lres) {
      lres = New refcounted<lookup3res>;
      call (NFSPROC3_LOOKUP, &argp->where, lres,
	    wrap (mkref (this), &cryptfs::reply_create,
		  nc, lres, ares), ao);
      return;
    }
    else if (lres->status) {
      nc->error (lres->status);
      return;
    }
    resp->resok->obj.set_present (true);
    *resp->resok->obj.handle = lres->resok->object;
  }

  if (!resp->resok->obj_attributes.present) {
    if (lres && lres->resok->obj_attributes.present)
      resp->resok->obj_attributes = lres->resok->obj_attributes;
    else if (!ares) {
      ares = New refcounted<getattr3res>;
      call (NFSPROC3_GETATTR, &*resp->resok->obj.handle, ares,
	    wrap (mkref (this), &cryptfs::reply_create,
		  nc, lres, ares), ao);
      return;
    }
    else if (ares->status) {
      nc->error (ares->status);
      return;
    }
    else {
      resp->resok->obj_attributes.set_present (true);
      *resp->resok->obj_attributes.attributes = *ares->attributes;
    }
  }

  if (!resp->resok->obj_attributes.attributes->size)
    vNew fhiv (this, *resp->resok->obj.handle,
	       &*resp->resok->obj_attributes.attributes);
  else
    resp->resok->obj_attributes.attributes->size
      = fixsize (resp->resok->obj_attributes.attributes->type,
		 resp->resok->obj_attributes.attributes->size);

  nc->reply (nc->getvoidres ());
}


void
cryptfs::reply_read (nfscall *nc, clnt_stat stat)
{
  if (stat) {
    reply (nc, stat);
    return;
  }
  read3res *res = nc->template getres<read3res> ();
  if (res->status || !res->resok->data.size ()) {
    reply (nc, stat);
    return;
  }

  read3args *arg = nc->template getarg<read3args> ();
  u_int64_t iv;
  switch (getiv (arg->file, false, nc->getaup (),
		 wrap (mkref (this), &cryptfs::reply_read, nc, stat), &iv)) {
  case 0:
    return;
  case -1:
    nc->error (NFS3ERR_IO);
    return;
  }

  rpc_bytes<RPC_INFINITY> data;
  swap (res->resok->data, data);
  char *p = data.base (), *e = p + (data.size () & ~blockmask);
  u_int64_t pos = arg->offset & ~blockmask;

  for (; p < e; p += blocksize, pos += blocksize)
    decrypt (pos, iv, p, p);

  u_int skip = arg->offset & blockmask;
  u_int count = data.size ();
  if (res->resok->eof && (count & blockmask))
    count -= blocksize;
  count -= skip;
  if (count > arg->count)
    count = arg->count;
  res->resok->data.set (data.base () + skip, count);
  res->resok->count = count;

  reply (nc, stat);
}

void
cryptfs::reply_write (nfscall *nc, u_int64_t end, clnt_stat stat)
{
#ifdef CFSASYNC
  if (!stat && !*nc->template getres<nfsstat3> ()
      && nc->proc () == NFSPROC3_WRITE)
    verf = nc->template getres<write3res> ()->resok->verf;
#endif /* CFSASYNC */
  if (!stat && !*nc->template getres<nfsstat3> ()
      && ((nc->proc () == NFSPROC3_WRITE
	   && nc->template getarg<write3args> ()->stable != UNSTABLE)
	  || nc->proc () == NFSPROC3_COMMIT))
    switch (getiv (*nc->getfh3arg (), true, nc->getaup (),
		   wrap (mkref (this), &cryptfs::reply, nc, stat))) {
    case -1:
      nc->error (NFS3ERR_IO);
      return;
    case 0:
      return;
    }

  /* Plaintext files that are not a multiple of 16 bytes in length
   * have 16 bytes added to ciphertext files.  Thus, while a 16 byte
   * file is only 16 bytes, a 15 byte file is actually 31 bytes.
   *
   * This causes a problem when, for instance, appending one byte to a
   * 15 byte file--the resulting ciphertext file will still be 31
   * bytes, so the plaintext file will still appear to be 15 bytes.
   *
   * We must correct the problem by truncating files in this situation.
   */
  if (nc->proc () == NFSPROC3_WRITE && !(end & blockmask)) {
    post_op_attr *poa
      = &nc->template getres<write3res> ()->resok->file_wcc.after;
    if (!poa->present)
      fillpoa (nc, nc->getfh3arg (), poa,
	       wrap (this, &cryptfs::reply_write, nc, end));
    else if (poa->attributes->size < end + blocksize
	     && poa->attributes->size > end) {
      setattr3args arg;
      arg.object = *nc->getfh3arg ();
      arg.new_attributes.size.set_set (true);
      *arg.new_attributes.size.val = end;
      arg.guard.set_check (true);
      *arg.guard.ctime = poa->attributes->ctime;
      static wccstat3 garbage;
      call (NFSPROC3_SETATTR, &arg, &garbage,
	    wrap (mkref (this), &cryptfs::reply, nc), ao);
      return;
    }
  }

  reply (nc, stat);
}

void
cryptfs::reply_write_read (nfscall *nc, ptr<read3res> rres, clnt_stat stat)
{
  if (stat) {
    reply (nc, stat);
    return;
  }
  if (rres && rres->status) {
    nc->template getres<write3res> ()->set_status (rres->status);
    reply (nc, stat);
    return;
  }

  write3args *argp = nc->template getarg<write3args> ();
  u_int64_t iv;
  switch (getiv (argp->file, false, nc->getaup (),
		 wrap (mkref (this), &cryptfs::reply_write_read,
		       nc, rres, stat),
		 &iv)) {
  case 0:
    return;
  case -1:
    nc->error (NFS3ERR_IO);
    return;
  }

  rpc_bytes<RPC_INFINITY> data;
  write3args arg = *argp;
  u_int64_t offset = arg.offset & ~blockmask;
  u_int32_t count = arg.count + (arg.offset & blockmask);
  if (count & blockmask) {
    if (rres->resok->eof)
      count += blocksize;
    else
      count = count + blockmask & ~blockmask;
  }

  if (rres) {
    data.setsize (count);
    bzero (data.base () + data.size () - blocksize, blocksize);
    if (rres->resok->data.size () >= blocksize) {
      decrypt (offset, iv, data.base (), rres->resok->data.base ());
      u_int32_t lastblk = (count & ~blockmask) - blocksize;
      if (lastblk + blocksize <= rres->resok->data.size ())
	decrypt (offset + lastblk, iv, data.base () + lastblk,
		 rres->resok->data.base () + lastblk);
    }
    memcpy (data.base () + (arg.offset - offset),
	    arg.data.base (), arg.data.size ());
    swap (data, arg.data);
    arg.offset = offset;
    arg.count = count;
  }
  assert (arg.offset == offset && arg.count == count);

  for (char *p = arg.data.base (),
	 *e = p + (arg.data.size () & ~blockmask);
       p < e; p += blocksize, offset += blocksize)
    encrypt (offset, iv, p, p);

  authopaque_set (ao, nc->getaup ());
  call (nc->proc (), &arg, nc->getvoidres (),
	wrap (mkref (this), &cryptfs::reply_write, nc,
	      argp->offset + argp->count), ao);
}

void
cryptfs::nfsdispatch (nfscall *nc)
{
  authopaque_set (ao, nc->getaup ());

  switch (nc->proc ()) {
  case NFSPROC3_LINK:
    fixdirop (&nc->template getarg<link3args> ()->link);
    break;
  case NFSPROC3_RENAME:
    fixdirop (&nc->template getarg<rename3args> ()->from);
    fixdirop (&nc->template getarg<rename3args> ()->to);
    break;
  case NFSPROC3_SYMLINK:
    {
      symlink3args *arg = nc->template getarg<symlink3args> ();
      fixdirop (&arg->where);
      u_int64_t h;
      rnd.getbytes (&h, sizeof (h));
      arg->symlink.symlink_data = encodename (0, armor64A (&h, 6)
					      << arg->symlink.symlink_data);
      break;
    }
  case NFSPROC3_LOOKUP:
  case NFSPROC3_MKDIR:
  case NFSPROC3_MKNOD:
  case NFSPROC3_REMOVE:
  case NFSPROC3_RMDIR:
    fixdirop (static_cast<diropargs3 *> (nc->getvoidarg ()));
    break;
  case NFSPROC3_SETATTR:
    {
      setattr3args *arg = nc->template getarg<setattr3args> ();
      if (arg->new_attributes.size.set) {
	*arg->new_attributes.size.val += datashift;
	if (*arg->new_attributes.size.val & blockmask)
	  *arg->new_attributes.size.val += blocksize;
      }
      // XXX - need to zero out the file if growing it
      break;
    }
  case NFSPROC3_CREATE:
    fixdirop (static_cast<diropargs3 *> (nc->getvoidarg ()));
    call (nc->proc (), nc->getvoidarg (), nc->getvoidres (),
	  wrap (mkref (this), &cryptfs::reply_create, nc,
		ptr<lookup3res> (NULL), ptr<getattr3res> (NULL)),
	  ao);
    return;
  case NFSPROC3_READ:
    {
      nc->template getarg<read3args> ()->offset += datashift;
      read3args arg = *nc->template getarg<read3args> ();
      getiv (arg.file, false, nc->getaup ());
      arg.count += arg.offset & blockmask;
      if (arg.count & blockmask)
	arg.count += blocksize;	// More than needed to decrypt, to get eof flag
      arg.offset &= ~blockmask;
      call (nc->proc (), &arg, nc->getvoidres (),
	    wrap (mkref (this), &cryptfs::reply_read, nc), ao);
      return;
    }
  case NFSPROC3_WRITE:
    {
      // XXX - need to zero out the file if writing beyond end
      write3args *arg = nc->template getarg<write3args> ();
#ifdef CFSASYNC
      arg->stable = UNSTABLE;
#endif /* CFSASYNC */
      getiv (arg->file, false, nc->getaup ());
      arg->offset += datashift;
      if (arg->offset & blockmask || arg->count & blockmask) {
	read3args rarg;
	rarg.file = arg->file;
	rarg.offset = arg->offset & ~blockmask;
	rarg.count = arg->count + (arg->offset & blockmask);
	if (rarg.count & blockmask)
	  rarg.count += blocksize; // To get eof flag in read
	ref<read3res> rres = New refcounted<read3res>;
	call (NFSPROC3_READ, &rarg, rres,
	      wrap (mkref (this), &cryptfs::reply_write_read, nc, rres),
	      ao);
      }
      else
	reply_write_read (nc, NULL, RPC_SUCCESS);
      return;
    }
  case NFSPROC3_COMMIT:
    {
#ifdef CFSASYNC
      commit3res res (NFS3_OK);
      res.resok->verf = verf;
      nc->reply (&res);
      return;
#endif /* CFSASYNC */
      commit3args *arg = nc->template getarg<commit3args> ();
      arg->offset += datashift;
      arg->count += arg->offset & blockmask;
      arg->count = arg->count + blockmask & blockmask;
      call (nc->proc (), nc->getvoidarg (), nc->getvoidres (),
	    wrap (mkref (this), &cryptfs::reply_write,
		  nc, arg->offset + arg->count), ao);
      return;
    }
  }

  call (nc->proc (), nc->getvoidarg (), nc->getvoidres (),
	wrap (mkref (this), &cryptfs::reply, nc), ao);
}

cryptfs::cryptfs (sfs_aid a, ref<nfsserv> s, clnt_t c, str pwd, int fd)
  : nfsc (c), nfss (s), fsfd (fd), aid (a)
{
  char key[sha1::hashsize];
  sha1_hash (key, pwd, pwd.len ());
  fskey.setkey (key, 16);
  bzero (key, sizeof (*key));
  nfss->setcb (wrap (this, &cryptfs::nfsdispatch));
}
